#
# Authorization based on {login, plain-text-pass} pair.
#
# Create password:
#   salt = random_256_bits_string()
#   hash = sha256(plain_pass + salt)
#   save {hash, salt} pair inside user thread data.
#
# Verification:
#   load {expectedHash, salt} pair from user thread data.
#   hash = sha256(plain_pass + salt)
#   compare hash with expectedHash
#

K = require('kcore')

K.Error.throwExceptionIfParameterNotSet(K.AuthProvider, 'K.AuthProvider')

crypto = require('crypto')

class K.AuthProviderPassword extends K.AuthProvider
  # Use 256-bit salt per each password.
  SALT_LEN_IN_BYTES = 32

  constructor: (kRequest) ->
    super(kRequest)

  #
  # ---------------------------------------------------------------------------
  #
  #                                Public API
  #
  # ---------------------------------------------------------------------------
  #

  getAuthenticatedUserModelByAuthId: (authId, done) ->
    @isAuthenticatedIdentity authId, (err, isAuth) =>
      if err?
        # Error while checking is user authorized.
        @appLog.logIfError(err)
        done(err)

      else if isAuth
        # User authorized. Get his model.
        @app.getModel(authId, done)

      else
        # User not authorized.
        #TODO: maybe some clever "else" ?
        #like checking if there is some other authenticated user we can select
        done(K.Error.createNotAuthenticatedByProviderError(authId, 'password'))

  getAuthenticatedIdentities: (done) ->
    @kRequest.getSessionForRead (err, session) =>
      if err?
        # Error while retrieving session.
        done(err)

      else
        # We got session object, read needed data from it.
        sha256Data = session.getOrDefault('auth.pass', {})
        session.release()
        done(null, Object.keys(sha256Data))

  isAuthenticatedIdentity: (authId, done) ->
    @kRequest.getSessionForRead (err, session) =>
      if err?
        # Error while retrieving session.
        done(err)

      else
        # We got session object, read needed data from it.
        sha256Data = session.getOrDefault('auth.pass', {})
        session.release()
        done(null, sha256Data[authId]?)

  setUserPasswordFromPlain: (userUid, passwd, done) ->
    # Generate salt.
    salt = @_getRandomString(SALT_LEN_IN_BYTES)

    # Compute hash(passwd + salt)
    hash = @_calculateHash(passwd + salt)

    # Set new {hash, salt} pair
    @_setUserPasswordInternal(userUid, hash, salt, done)

  setUserPasswordFromAlreadyPreparedHash: (userUid, hash, done) ->
    K.Error.throwNotImplemented('K.AuthProviderPassword::setUserPasswordFromArleadyPreparedHash')

  # Verify is given password match the password passed in
  # setPasswordXXX() method before.
  # done = (err, result)
  checkPassword: (userUid, passwd, done) ->
    rv = false
    @app.getModel userUid, (err, userThread) =>
      if err?
        # Error while loading user thread.
        @appLog.logIfError(err)
        done(err)
      else
        # User thread loaded. Get {expectedHash, salt} pair.
        actionObj =
          cmd: 'getAppData'
          caller: 'system'
          app: 'authPassword'

        userThread.action actionObj, (err, appData) =>
          if err?
            # Error while loading application data from user thread.
            @appLog.logIfError(err)
            done(err)
          else
            # Application data loaded. We should get {hash, salt} pair
            # saved in _setUserPasswordInternal() before.
            if appData?
              expectedHash = appData['sha256Hash']
              salt         = appData['sha256Salt']

              if expectedHash? and salt?
                # Compute sha256(passwd + salt).
                # We should get the same hash as expectedHash loaded
                # from user thread.
                hash = @_calculateHash(passwd + salt)
                rv   = (hash is expectedHash)

            # Pass verification result to caller.
            done(null, rv)

  # Check is given password correct and update session if yes.
  # done = (err, result)
  authenticate: (userUid, passwd, done) ->
    @checkPassword userUid, passwd, (err, result) =>
      if err?
        # Unexpecting error while checking password.
        done(err)
      else
        if result is true
          # Password is correct, modify session.
          @kRequest.getSessionForWrite (err, session) =>
            if err?
              # Error while retrieving session.
              done(err)

            else
              # We got session object, go on.
              sha256Data = session.getOrDefault('auth.pass', {})

              # TODO: Add login time.
              sha256Data[userUid] = {uid: userUid}
              session.setData('auth.pass', sha256Data)

              # Flush session to db/disk.
              session.release (err) =>
                if err?
                  # Error while flushing session data to storage.
                else
                  # Session flushed.
                  done(null, 'ok')
        else
          # Wrong password.
          done(null, 'youShallNotPass')

  deauthenticate: (userUid, done) ->
    # Password is correct, modify session.
    @kRequest.getSessionForWrite (err, session) =>
      if err?
        # Error while retrieving session.
        done(err)

      else
        # We got session object, go on.
        sha256Data = session.getOrDefault('auth.pass', {})

        if sha256Data[userUid]?
          delete sha256Data[userUid]
          session.setData('auth.pass', sha256Data)

        session.release(done)

  deauthenticateAll: (done) ->
    # Password is correct, modify session.
    @kRequest.getSessionForWrite (err, session) =>
      if err?
        # Error while retrieving session.
        done(err)

      else
        # We got session object, go on.
        session.setData('auth.pass', {})
        session.release(done)

  #
  # ---------------------------------------------------------------------------
  #
  #                             Internal helpers
  #
  # ---------------------------------------------------------------------------
  #

  _setUserPasswordInternal: (userUid, hash, salt, done) ->
    @app.getModel userUid, (err, userThread) =>
      if err?
        # Error while retrieving user thread.
        @appLog.logIfError(err)
        done(err)
      else
        actionObj =
          cmd: 'setAppData'
          caller: 'system'
          app: 'authPassword'
          appData: {sha256Hash: hash, sha256Salt: salt}

        userThread.action actionObj, (err) =>
          @appLog.logIfError(err)
          done(err)

  _getRandomString: (lengthInBytes) ->
    # Generate raw buffer.
    rv   = ''
    blob = crypto.randomBytes(lengthInBytes)

    # Convert raw buffer into hex string.
    for byte in blob
      if byte <= 0x0f
        rv += '0' + byte.toString(16)
      else
        rv += byte.toString(16)

    return rv

  _calculateHash: (data) ->
    return crypto.createHash('sha256').update(data).digest("hex")
