K = require('kcore')

class K.EventServer
  constructor: (@app) ->
    K.Error.throwExceptionIfParameterNotSet(@app, 'app')

    @watchers    = {}
    @watchedUids = {}

  _compareStamps: (kRequest, arrayOfItems) ->
    rv = []

    for item in arrayOfItems
      status      = 'internalError'
      sync        = 'unknown'
      serverStamp = null
      statusHint  = null

      # Check parameters.
      if not item.uid?
        status     = 'missingParameter'
        statusHint = 'missing uid parameter'

      else if not item.stamp?
        status     = 'missingParameter'
        statusHint = 'missing stamp parameter'

      else if not K.UidUtil.isOkAsUid(item.uid)
        status     = 'invalidParameter'
        statusHint = 'uid parameter seems not correctly'

      else
        # Parameters seems ok. Check sync status.
        status = 'ok'

        if @watchedUids[item.uid]?
          # Uid found, check is caller stamp up to date.
          serverStamp = @watchedUids[item.uid].stamp

          if not serverStamp?
            # We don't know stamp yet.
            sync = 'unknown'

          else if serverStamp is item.stamp
            # Caller stamp is up to date.
            sync = 'sync'

          else
            # Caller stamp is out of date.
            sync = 'changed'

      # Accumulate results.
      newResultItem = {
        uid: item.uid
        status: status
        sync: sync
      }

      rv.push(newResultItem)

      if statusHint?
        # Avoid adding null statusHint
        newResultItem.statusHint = statusHint

      if serverStamp?
        # Avoid adding null serverStamp
        newResultItem.serverStamp = serverStamp

    # Return result to caller.
    return rv

  #
  # Add array of {uid, stamp} pairs to watch list within given kRequest context.
  # After one of uid changed its state via notifyEvent() function, event server
  # will  sent notification to related kRequest object.
  #
  # There are two working modes:
  #
  # - ajax based: one way request only (browser asks server), we check are given
  #   uids in sync (have up to date stamp value) and return result immediately.
  #
  # - websocket based: two way requests available, remember uids to be watched
  #   and send notifiaction to browser if of of uid changed its stamp.
  #
  # arrayOfItems = [{uid: 'item.1', stamp: '20'},
  #                 {uid: 'item.2', stamp: '10'},
  #                 {uid: 'item.3', stamp: '15'},
  #                 ...
  #                ]
  #
  watch: (kRequest, arrayOfItems, done) ->
    switch kRequest.getMethod()
      #
      # Ajax based.
      # Just compare stamps and return result immediately.
      #
      when 'ajax'
        rv = @_compareStamps(kRequest, arrayOfItems)
        done(null, rv)

      #
      # Websocket based.
      # Remember uids to be watched and send notification via websocket if
      # one of uids changed its stamp.
      #
      when 'ws'
        # First, compare all stamps, which we're going to watch like in ajax case.
        rv = @_compareStamps(kRequest, arrayOfItems)

        # Unwatch any current watched uids if exist (within kRequest context).
        @unwatchAllByKRequest(kRequest)

        # Set up new watched uids list from scratch.
        newWatcher = {uids: [], kRequest: kRequest}
        wsId       = kRequest.getWebsocket().id

        for item in rv
          if item.status is 'ok'

            # Collect uids watched by current watcher.
            # This allow us to do quick search by kRequest key.
            newWatcher.uids.push(item.uid)

            # Callect watchers tracking given uid.
            # This allow us to do quick search by uid.
            if not @watchedUids[item.uid]?
              @watchedUids[item.uid] = {watchers: {}}

            @watchedUids[item.uid].watchers[wsId] = newWatcher

            # Change status to 'watching'
            item.status = 'watching'

        # Add new watcher, but only if at least one uid added to watch list.
        if newWatcher.uids.length > 0
          @watchers[wsId] = newWatcher

        # Pass result to caller.
        done(null, rv)

      #
      # Fatal, should never exists.
      #
      else
        throw Error({errorCode: 'UnknownRequestMethod', ext:{method: kRequest.getMethod()}})


  unwatchAllByKRequest: (kRequest) ->
    wsId = kRequest.getWebsocket().id

    if @watchers[wsId]?
      # Detach watcher from all uids.
      for uid in @watchers[wsId].uids
        if @watchedUids[uid]?
          delete @watchedUids[uid].watchers[wsId]

      # Unregister watcher.
      delete @watchers[wsId]

  #
  # Notify EventServer, that one of watched uid changed its stamp. This call
  # should triggers sending notification to browser.
  #
  notifyEvent: (uid, stamp) ->
    if not @watchedUids[uid]?
      @watchedUids[uid] = {watchers: {}}

    if @watchedUids[uid].stamp isnt stamp
      # Stamp changed. Save new stamp.
      @watchedUids[uid].stamp = stamp

      # Notify watching kRequests if any.
      if @watchedUids[uid].watchers?
        for wsId, watcher of @watchedUids[uid].watchers
          event = JSON.stringify({
            channelId:'EventServer',
            event:{uid: uid, serverStamp: stamp}})
          watcher.kRequest.getWebsocket().send(event)

  onKRequestClosed: (kRequest) ->
    if kRequest.getMethod() is 'ws'
      @unwatchAllByKRequest(kRequest)

  reset: () ->
    # WARNING: This is test/development call.
    # Not to be used in production environment.
    @watchers    = {}
    @watchedUids = {}

  #
  # WARNING: Below functions are test/development only.
  # Not to be used in production environment.
  #

  getTotalNumberOfWatchers: () ->
    return Object.keys(@watchers).length

  getNumberOfWatchersByUid: (uid) ->
    if not @watchedUids[uid]?
      return 0
    else
      return Object.keys(@watchedUids[uid].watchers).length

  getArrayOfWatchedUids: () ->
    rv = []
    for uid, item of @watchedUids
      if Object.keys(item.watchers).length > 0
        rv.push(uid)

    return rv
