K      = require('kcore')
fs     = require('fs-extra')
coffee = require('coffee-script')
vm     = require('vm')

K.Error.throwExceptionIfParameterNotSet(K.LockManager, 'K.LockManager')

class K.FileManager
  constructor: (app) ->
    # Check parameters.
    K.Error.throwExceptionIfParameterNotSet(app, 'app')
    K.Error.throwExceptionIfParameterNotSet(app.getLog(), 'appLog')
    K.Error.throwExceptionIfParameterNotSet(app.getLockManager(), 'lockManager')

    # Init object.
    @app          = app
    @appLog       = app.getLog()
    @lockManager  = app.getLockManager()


  #
  # Lock functions.
  #
  openExclusiveAccessToClass: (done) ->
    # Lock whole K.FileManager class i.e. all object instances and static
    # data if any.
    @lockManager.openExclusiveAccess('K.FileManager', done)

  openExclusiveAccessToObject: (done) ->
    # Lock this object.
    # TODO: Not implemented.
    done(K.Error.createNotImplementedError('K.FileManager::openExclusiveAccessToObject'))

  openExclusiveAccessToPath: (path, done) ->
    # Lock selected path on disk only eg. 'data/common/users.txt' file.
    normalizedPath = K.FileUtil.normalizePath(path)

    if not normalizedPath?
      # Error while noralizing path.
      err = K.Error.createInvalidParameterError('path', path)
      @appLog.logIfError(err)
      done(err)
    else
      # Open exclusive access on normalized path.
      @lockManager.openExclusiveAccess(normalizedPath, done)

  #
  # Non-atomic API.
  # Caller MUSTS use own lock mechanism if needed eg.
  #
  # lockManager.exclusiveAccess 'file1', (lockHandle) =>
  #   fileManager.countLines(...)
  #   fs.appendFile(...)
  #   fs.doSomething(...)
  #   lockHandle.close()
  #

  # The throwWhenFailed parameter is a "level of trust":
  #  should be true when we know the file should be ALWAYS there (for example for asset files),
  #   so when the fail is not there - whole code throws exception.
  #  When we're not sure if the file is there (e.g. dynamic/user file) - then use false here.
  loadDataFromJson: (path, throwWhenFailed, done) ->
    fs.readFile path, 'utf8', (err, data) =>
      if not err?
        # File loaded, parse JSON data
        try
          err = null
          plainData = JSON.parse(data.toString())
        catch jsonError
          # Error while parsing
          err = new K.Error({errorCode: 'jsonParseError', ext: {path:path, jsonError:jsonError}})
          if throwWhenFailed
            throw new Error('cannot json-parse from file in path:' + path)
          plainData = ''
          @appLog.log(err.msg)
        # return the result
        done err, plainData
      else
        if throwWhenFailed
          throw new Error('cannot load from file in path:' + path)
        else
          done err

  loadDataFromCoffee: (path, done) ->
    fs.readFile path, 'utf8', (err, data) =>
      if not err?
        compiled = coffee.compile(data)
        sandbox = {rv:{}}
        vm.runInNewContext(compiled, sandbox)
        done null, sandbox.rv
      else
        console.log 'error loading data from coffee:', path
        throw new Error('cannot load')

  countLines: (filepath, done) ->
    # Count number of lines in file.
    numberOfLinesCnt = 0
    currentLineSize  = 0
    fs.createReadStream(filepath)
      .on 'data', (chunk) =>
        # New piece of data readed, count number of end of line inside.
        for idx in [0...chunk.length]
          # We're inside line, go on.
          currentLineSize++
          if currentLineSize is 1
            # Non-empty line found (including EOL character), increase counter.
            numberOfLinesCnt++

          if chunk[idx] is 10
            # Line break, reset counter.
            currentLineSize = 0

      .on 'end', () =>
        # End of file, pass result to caller.
        done(null, numberOfLinesCnt)

      .once 'error', (err) =>
        # Error occures while reading.
        done(err)

  # Write data to file, but only if it doesn't exist yet.
  writeIfNotExists: (filepath, data, done) ->
    fs.exists filepath, (exists) =>
      if exists
        # Error, file already exits.
        done?(K.Error.createFileAlreadyExistsError(filepath))
      else
        # Model does not exists yet. Init and save new one.
        fs.writeFile filepath, data, (ioError) =>
          if ioError?
            # Error while creating file.
            @appLog.logIfError(ioError)
            done?(new K.Error({errorCode: 'ioError', ext: {ioError: ioError}}))

          else
            # Success, data written.
            done?(null)

  #
  # Atomic API.
  # These functions call app.getLockManager() to setup the lock on the given filepath
  #
  # WARNING: Caller should NOT lock the file by his own due to possible deadlock.
  # If you need to do serie of operations on single file,
  # use non-atomic API with you own explicit lock calls.
  #

  atomicCountLines: (filepath, done) ->
    @openExclusiveAccessToPath filepath, (lockError, lockHandle) =>
      if lockError?
        done(lockError)
      else
        @countLines(filepath, done)
        lockHandle.close()

  atomicWriteIfNotExists: (filepath, data, done) ->
    @openExclusiveAccessToPath filepath, (lockError, lockHandle) =>
      if lockError?
        done(lockError)
      else
        @writeIfNotExists(filepath, data, done)
        lockHandle.close()

  atomicWrite: (filepath, data, done) ->
    @openExclusiveAccessToPath filepath, (lockError, lockHandle) =>
      if lockError?
        done(lockError)
      else
        # Internal wrapper for done callback.
        # Close exclusive access and log error if needed.
        _done = (extError) =>
          lockHandle.close()
          if extError?
            err = new K.Error({errorCode: 'atomicWriteError', ext: extError})
            @appLog.logIfError(err)
            done?(err)
          else
            done(null)

        # Check parameters.
        if not filepath?
          _done?(K.Error.createParameterNotSetError('filepath'))
        else if not data?
          _done?(K.Error.createParameterNotSetError('data'))
        else
          # Write data to lock file first.
          lockFilepath = filepath + '.lock'

          fs.writeFile lockFilepath, data, (writeError) =>
            if writeError?
              # Error while writing to lock file.
              _done?(writeError)
            else
              # Lock file saved, rename it to destination file.
              fs.rename lockFilepath, filepath, (renameError) =>
                if renameError?
                  # Error while renaming file.
                  _done?(renameError)
                else
                  # All OK, pass success to caller.
                  _done?(null)

  atomicAppend: (filepath, data, done) ->
    @openExclusiveAccessToPath filepath, (lockError, lockHandle) =>
      if lockError?
        done(lockError)
      else
        # Internal wrapper for done callback.
        # Close lock handle and log error if needed.
        _done = (extError) =>
          lockHandle.close()
          if extError?
            err = new K.Error({errorCode: 'atomicAppendError', ext: extError})
            @appLog.logIfError(err)
            done?(err)
          else
            done?(null)

        # Check parameters.
        if not filepath?
          _done?(K.Error.createParameterNotSetError('filepath'))
        else if not data?
          _done?(K.Error.createParameterNotSetError('data'))
        else if data is ''
          # Nothing to append.
          _done?(null)
        else
          # Copy original lock file first.
          lockFilepath = filepath + '.lock'

          fs.copy filepath, lockFilepath, (copyError) =>
            if copyError?
              # Error while copying file.
              _done(copyError)
            else
              # File copied, append data to lock file.
              fs.appendFile lockFilepath, data, (appendError) =>
                if appendError?
                  # Error while appending data.
                  _done?(appendError)
                else
                  # Data appended to lock file. Rename lock to original.
                  fs.rename lockFilepath, filepath, (renameError) =>
                    if renameError?
                      # Error while renaming file.
                      _done?(renameError)
                    else
                      # All OK, pass success to caller.
                      _done(null)
