K    = require('kcore')
fs   = require('fs-extra')
os   = require('os')
path = require('path');

jsonStringifySafe = require('json-stringify-safe')

require('./KLogCollector.coffee')

# -----------------------------------------------------------------------------
#                            Log transport classes
# -----------------------------------------------------------------------------

class K.LogTransport
  pushEntry: (entry) ->
    K.Error.throwNotImplemented('K.LogTransport::pushEntry')

  getLogCollector: () ->
    return @logCollector

class K.LogTransportNative extends K.LogTransport
  constructor: () ->
    super(filePath)

class K.LogTransportToTxtFile extends K.LogTransport
  constructor: (filePath) ->
    super()

    @filePath   = filePath
    @logDir     = path.dirname(@filePath)
    @rotateSize = 100 * 1024 * 1024

    # Make sure that log directory exists.
    fs.mkdirpSync(@logDir)

    if fs.existsSync(filePath)
      # Log file already exists. Continue it.
      @startTime = @_getStartTimeFromFile(filePath)
      @logSize   = fs.statSync(filePath).size
      @fd        = fs.openSync(filePath, 'a')

    else
      # Create new log file from stratch.
      @_initNewLogFile()

    @logCollector = new K.LogCollectorFromJsonFile(@)

  #
  # Public API.
  #

  getFilePath: () ->
    return @filePath

  getCurrentStartTime: () ->
    return @startTime

  pushEntry: (entry, forceFlush = false) ->
    if forceFlush
      fs.writeSync @fd, entry

    else
      fs.write @fd, entry, () ->
        # Empty done callback to make node 8.4.0 happy.
        # Calling async functions without callback is not legal anymore.
        ;

    # Limit logfile size.
    @logSize += entry.length
    if @logSize > @rotateSize
      @_rotateLogFile()

  #
  # Internal helpers.
  #

  _getTimestampMs: () ->
    return (new Date()).getTime()

  _getStartTimeFromFile: (filePath) ->
    # Read first 64 bytes from existing log file.
    buffer = new Buffer(64)
    fd     = fs.openSync(filePath, 'r')

    fs.readSync(fd, buffer, 0, 64, 0)
    fs.closeSync(fd)

    # Isolate first line only.
    lines = buffer.toString().split("\n")

    # We're expecting startTime on the first line. Parse it.
    headEntry = JSON.parse(lines[0])

    if (headEntry?[0]?[0] is 'logfile') and (headEntry?[1]?.started?)
      rv = headEntry[1].started
    else
      # Corrupted log. Move to corrupted directory and start new one from stratch.
      @_rotateCorruptedLogFile()
      rv = @startTime

    return rv

  _initNewLogFile: () ->
    @startTime = @_getTimestampMs()
    @logSize   = 0
    @fd        = fs.openSync(@filePath, 'w')
    @pushEntry([['logfile'], {started: @startTime}], true)

  _rotateLogFile: () ->
    fs.closeSync(@fd)
    fs.renameSync(@filePath, @filePath + '.' + @startTime)
    @_initNewLogFile()

  _rotateCorruptedLogFile: () ->
    rnd           = parseInt(Math.random() * 1000000000)
    logDir        = require('path').dirname(@filePath)
    fileName      = require('path').basename(@filePath)
    corruptedDir  = logDir + '/corrupted'
    corruptedPath = corruptedDir + '/corrupted_' + fileName + '.' + rnd

    if @fd?
      fs.closeSync(@fd)

    fs.mkdirpSync(corruptedDir)
    fs.renameSync(@filePath, corruptedPath)
    @_initNewLogFile()

class K.LogTransportToJsonFile extends K.LogTransportToTxtFile
  constructor: (filePath) ->
    super(filePath)

  pushEntry: (entry) ->
    super(jsonStringifySafe(entry) + '\n')

# -----------------------------------------------------------------------------
#                             Log context classes
# -----------------------------------------------------------------------------

class K.LogCtx
  constructor: (@transports, @ctxStamp) ->
    @subLogId = 0

  isFake: () ->
    return false

  # returns stamp to use for sublog
  createSubLog: (subLogInfo) ->
    @subLogId++
    @log({'new':@subLogId, 'info':subLogInfo})
    newStamp = @ctxStamp.slice() # array copy
    newStamp.push @subLogId
    return newStamp

  getCtxStamp: () ->
    return @ctxStamp

  getTransports: () ->
    return @transports

  pushEntry: (entry) ->
    @transports[0].pushEntry([@ctxStamp, entry])

  log: (msg...) ->
    args = Array.prototype.slice.call(arguments)
    if args.length > 1
      args = [args]
    @pushEntry.apply @, args

  logIfError: (err) ->
    if err?
      @logTime {eventType:'error', err:err}

  logTime: (more...) ->
    args = Array.prototype.slice.call(arguments)
    @pushEntry {'time':(new Date()).getTime(), 'event':args}

class K.LogCtxFake
  constructor: (@transports, @ctxStamp) ->
    ;

  isFake: () ->
    return true

  createSubLog: (subLogInfo) ->
    return null

  getTransports: () ->
    return null

  log: (msg...) ->
    ;

  logIfError: (err) ->
    ;

  logTime: (more...) ->
    ;

class K.LogCtxWorkspace extends K.LogCtx
  constructor: (@transports, srvId) ->
    super(@transports, [srvId])
    @pushDefaults()

  pushDefaults: () ->
    @log('Starting...')
    @logTime('Server starting')
    @log({'os.hostname': os.hostname()})
    @log({'os.type':     os.type()})
    @log({'os.platform': os.platform()})
    @log({'os.arch':     os.arch()})
    @log({'os.totalmem': os.totalmem()})
    @log({'os.freemem':  os.freemem()})
    @log({'os.arch':     os.arch()})
    @log({'process.cwd': process.cwd()})

class K.LogCtxApp extends K.LogCtx
  constructor: (@parentServerLog, @app) ->
    ctxStamp = @parentServerLog.createSubLog({type:'app', appName: @app.getName()})
    super(@parentServerLog.getTransports(), ctxStamp)

class K.LogCtxRequest extends K.LogCtx
  constructor: (@parentAppLog, @req) ->
    ctxStamp = @parentAppLog.createSubLog({type:'req'})
    super(@parentAppLog.getTransports(), ctxStamp)

  logRequest: () ->
    @logTime()
    @req.log.log {
      httpVersionMajor: @req.httpVersionMajor
      httpVersionMinor: @req.httpVersionMinor
      upgrade: @req.upgrade
      socketBytesRead: @req.socket.bytesRead
      headers: @req.headers
      method: @req.method
      remoteAddress: @req.connection.remoteAddress
    }
