K = require('kcore')
glob = require("glob")
LineByLineReader = require('line-by-line')

class K.LogCollector
  getStartAddress: (done) ->
    K.Error.throwNotImplemented('K.LogCollector::getStartAddress')

  getEntries: (startAddress, length, done) ->
    K.Error.throwNotImplemented('K.LogCollector::getEntries')

  getEntriesAfter: (startAddress, length, done) ->
    K.Error.throwNotImplemented('K.LogCollector::getEntriesAfter')

class K.LogCollectorFromTxtFile extends K.LogCollector
  constructor: (@logTransport) ->
    super()
    @filePath = @logTransport.getFilePath()

  #
  # Public API.
  #

  getStartAddress: (done) ->
    @_getArchivedStartTimes (err, arrayOfStartTimes) =>
      if err?
        # Error while reading list of available archive files.
        done(err)

      else
        # List of archive files loaded. Get oldest one.
        rv = arrayOfStartTimes[0] + '.0'
        done(null, rv)

  getEntries: (startAddress, length, done) ->
    @_getEntriesInternal(startAddress, length, false, done)

  getEntriesAfter: (startAddress, length, done) ->
    @_getEntriesInternal(startAddress, length, true, done)

  #
  # Internal helpers.
  #

  _getTimestampMs: () ->
    return (new Date()).getTime()

  _decodeEntryFromString: (entry) ->
    return entry

  _getArchivedStartTimes: (done) ->
    rv = []

    glob @filePath + '.*', (err, files) =>
      if err?
        # Error while listing archive files.
        done(err)

      else
        # Archive file listed. Get start times from file names.
        for file in files
          explodedFile = file.split('.')
          startTime    = explodedFile[explodedFile.length - 1]
          rv.push(parseInt(startTime))

        # Add current file too.
        rv.push(@logTransport.getCurrentStartTime())

        # Success, pass array of available start times to caller.
        done(null, rv)

  _readArchivedEntries: (rv, startTime, startEntryIdx, length, done) ->
    filePath = @filePath
    lineIdx  = 0

    if startTime? and (startTime isnt @logTransport.getCurrentStartTime())
      # Add .timestamp postfix if it's not current logfile.
      filePath += '.' + startTime

    # Process archive file line by line.
    options = {
      encoding: 'utf8'
      skipEmptyLines: true
    }
    lr = new LineByLineReader(filePath, options)

    lr.on 'error', (err) =>
      # Error while reading archive file.
      done(err)

    lr.on 'end', () =>
      # Whole file processed. Work finished.
      done(null)

    lr.on 'line', (line) =>
      # Next line readed.
      if (lineIdx >= startEntryIdx) and (length > 0)
        rv.push {
          address: startTime + '.' + lineIdx
          entry:  @_decodeEntryFromString(line.trim())
        }
        length--

        if length <= 0
         # All data readed. Stop processing file here.
         lr.close()

      lineIdx++

  _getEntriesInternal: (startAddress, length, skipFirstEntry, done) ->
    rv = []

    # Split address into startTime and entryIdx.
    explodedAddress = startAddress.split('.')

    if explodedAddress.length isnt 2
      # Error - wrong address.
      done new K.Error {errorCode: 'wrongLogAddress', ext: {startAddress: startAddress}}

    else
      # Address decoded, go on.
      startTime = parseInt(explodedAddress[0])
      entryIdx  = parseInt(explodedAddress[1])

      # Get list of available archive files.
      @_getArchivedStartTimes (err, arrayOfStartTimes) =>
        if err?
          # Error while listing archive files.
          done(err)

        else
          # Archive files listed. Localize requested start time.
          startTimeIdx = arrayOfStartTimes.indexOf(startTime)

          if startTimeIdx is -1
            # Start time not found. Maybe old log file was deleted on server side?
            # Find nearsest one.
            for idx, time of arrayOfStartTimes
              if time > startTime
                startTimeIdx = idx
                startTime    = time
                entryIdx     = 0
                break

          else if skipFirstEntry
            entryIdx++

          if startTimeIdx is -1
            # Error - bad start time.
            done new K.Error {errorCode: 'wrongLogAddress', ext: {startAddress: startAddress}}

          else
            # Start time found, go on.
            _readNextArchiveOrDone = (startTimeIdx, entryIdx) =>
              if startTimeIdx >= arrayOfStartTimes.length
                # No more data available. Return what we have.
                done(null, rv)

              else if rv.length < length
                # We still need more entires. Read next archive file.
                startTime        = arrayOfStartTimes[startTimeIdx]
                entriesNeededCnt = length - rv.length

                @_readArchivedEntries rv, startTime, entryIdx, entriesNeededCnt, (err) =>
                  # Go to next archive in chain.
                  _readNextArchiveOrDone(startTimeIdx + 1, 0)

              else
                # We got needed amount of entries. Don't go on any more.
                done(null, rv)

            # Read first archive in chain.
            _readNextArchiveOrDone(startTimeIdx, entryIdx)

class K.LogCollectorFromJsonFile extends K.LogCollectorFromTxtFile
