K = require('kcore')

K.Error.throwExceptionIfParameterNotSet(K.Auth, 'K.Auth')
K.Error.throwExceptionIfParameterNotSet(K.AuthProviderPassword, 'K.AuthProviderPassword')
K.Error.throwExceptionIfParameterNotSet(K.AuthProviderAnonymous, 'K.AuthProviderAnonymous')

class K.Request
  @ANONYMOUS_NICK = 'anonymous'

  constructor: (@app) ->
    @expressRequest  = null
    @expressResponse = null
    @userNick        = K.Request.ANONYMOUS_NICK
    @kemuSid         = null
    @cookies         = null
    @sessionManager  = @app.getSessionManager()
    @auth            = null
    @appLog          = @app.getLog()
    @authProviders   = {}
    @method          = 'unknown'
    @ws              = null

  _parseCookieString: (cookieString) ->
    if cookieString?
      @cookies = {}

      tokens = cookieString.split(';')
      tokens.forEach (cookie) =>
        parts = cookie.split('=')
        @cookies[parts.shift().trim()] = decodeURI(parts.join('='))

  setResponseConfig: (responseConfig, done) ->
    @responseConfig = responseConfig

    # default address generation
    if @responseConfig.port != 80
      hostAndPort = @responseConfig.host + ':' + @responseConfig.port
    else
      hostAndPort = @responseConfig.host

    if not @responseConfig.rootLink?
      @responseConfig.rootLink = 'http://' + hostAndPort

    if not @responseConfig.wsLink?
      @responseConfig.wsLink = 'ws://' + hostAndPort

    if not @responseConfig.jsonLink?
      @responseConfig.jsonLink = 'http://' + hostAndPort + '/jsonActions'

    done(null)

  attachExpressRequestAndResponse: (expressReq, expressRes, done) ->
    @expressRequest  = expressReq
    @expressResponse = expressRes
    @_parseCookieString(@expressRequest.headers.cookie)

    # Create K.Auth object.
    @auth = new K.Auth(@)
    @method = 'ajax'

    done(null)

  attachWebsocket: (ws) ->
    @ws     = ws
    @method = 'ws'

  getApp: () ->
    return @app

  getMethod: () ->
    return @method

  getWebsocket: () ->
    return @ws

  getLog: () ->
    return @expressRequest.log

  getExpressRequest: () ->
    return @expressRequest

  getExpressResponse: () ->
    return @expressResponse

  getRootLink: (linkin) ->
    if linkin? and (linkin isnt '')
      return @responseConfig.rootLink + '/' + linkin
    else
      return @responseConfig.rootLink

  getJsonLink: () ->
    return @responseConfig.jsonLink

  getWsLink: () ->
    return @responseConfig.wsLink

  getUserNick: () ->
    # TODO
    K.Error.throwNotImplemented('K.Request::getUserNick')
    #return @userNick

  getKemuSid: () ->
    # TODO: We changed kemuSid to sid to avoid collisin with
    # PHP backed. Change back to kemuSid after php server dropped.
    if (not @kemuSid?) and @cookies?['sid']?
      @kemuSid = @cookies['sid']

    return @kemuSid

  getCookie: (name) ->
    rv = null
    if @cookies? and @cookies[name]?
      rv = @cookies[name]

    return rv

  # Usage:
  #   getSessionForXXX (err, session) =>
  #     session.setData(...)
  #     session.release() or session.discard()
  #
  _getSession: (intent, done) ->
    # Get session if parent application has related session manager.
    if @sessionManager?
      # Determine which function use to get session depending on
      # info does caller want to write or read only.
      getSessionCb = null
      if intent is 'read'
        getSessionCb = @sessionManager.getSessionForRead.bind(@sessionManager)
      else
        getSessionCb = @sessionManager.getSessionForWrite.bind(@sessionManager)

      # Get session from manager.
      # TODO: Remove debugs.
      @app.logDebug 'Retrieving session for', @getKemuSid(), intent

      getSessionCb @getKemuSid(), (err, session) =>
        if (not err?) and session?
          @app.logDebug 'SESSION DATA:', session.data
          # We got session object from manager.
          # Pass it to caller.
          @app.logDebug 'Got session', session.getId()
          @kemuSid = session.getId()
          done(null, session)

        else
          # Error while retrieving session.
          @app.logDebug 'ERROR: Cannot to get session', @getKemuSid(), intent
          done(err)
    else
      # Application has no support for sessions.
      err = new K.Error({'sessionManagerNotFound', ext:{}})
      @appLog.logIfError(err)
      done(err)

  getSessionForRead: (done) ->
     @_getSession('read', done)

  getSessionForWrite: (done) ->
     @_getSession('write', done)

  getAuth: () ->
    return @auth

  getApp: () ->
    return @app

  getAppLog: () ->
    return @appLog

  getAuthProvider: (id) ->
    rv = null

    if @authProviders[id]?
      # Provider object already created. Just return it.
      rv = @authProviders[id]

    else
      # Create new provider object.
      switch id
        when 'password'
          rv = new K.AuthProviderPassword(@)

        when 'default', 'anonymous'
          rv = new K.AuthProviderAnonymous(@)

        else
          # Unknown provider.
          err = new K.Error({errorCode: 'unknownAuthProviderId', ext:{id: id}})
          @appLog.logIfError(err)
          throw err

      # Cache provider object for further calls.
      @authProviders[id] = rv

    return rv

  getCurrentIdentity: (done) ->
    @getAuth().getCurrentUserModel (err, userThread) =>
      if err?
        # Unexpected error while retrieving current logged user.
        done(err)

      else
        # User thread retrived. Pass userUid to caller.
        done(null, userThread.getUid())

  #
  # Is any non-anonymous auth currently on.
  #
  # USAGE EXAMPLE: Check if the "anonymous nick" field entry should be presented.
  #
  isNonAnonymousAuthPresent: (done) ->
    @getAuth().getIdentitiesOfClass 'password', (err, identities) =>
      if err?
        done(err)
      else
        done(null, identities.length > 0)
