K = require('kcore')
express = require('express')
bodyParser = require('body-parser')
async = require('async')
compression = require('compression')
connectRestreamer = require('connect-restreamer')
K.requireLib('knate')
require('terminal-colors')

K.Error.throwExceptionIfParameterNotSet(K.Auth, 'K.Auth')
K.Error.throwExceptionIfParameterNotSet(K.AuthProvider, 'K.AuthProvider')
K.Error.throwExceptionIfParameterNotSet(K.EventServer, 'K.EventServer')
K.Error.throwExceptionIfParameterNotSet(K.FileManager, 'K.FileManager')
K.Error.throwExceptionIfParameterNotSet(K.LockManager, 'K.LockManager')

class K.ServerApp
  NICK_MAX_LEN         = 100
  IS_OK_AS_NICK_REGEXP = new RegExp('^[a-zA-Z0-9_]+$', '')

  # Email validation based on:
  # http://www.regular-expressions.info/email.html
  IS_OK_AS_EMAIL_REGEXP = new RegExp('^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,}$', '')

  PASSWORD_MAX_LEN = 64
  PASSWORD_MIN_LEN = 5

  constructor: (@appName, @config) ->
    if not @config?
      @config = {}

    @serverStartTime = null
    @serverLog       = null
    @appLog          = null
    @models          = {}
    @serverWorkspace = @config.serverWorkspace

    # log
    if @config.serverLog
      @serverLog = @config.serverLog
      @serverLog.log({'Starting app':@appName})
      @appLog    = new K.LogCtxApp(@serverLog, @)
    else
      # no log, so attach empty call
      @appLog = new K.LogCtxFake()

    @serverStats = new K.ServerStatsCollector()

    # default host
    if not @config.host?
      @config.host = '127.0.0.1'

    # default port
    if not @config.port?
      @config.port = 80

    # default address generation
    @defaultResponseConfig = {
      host:   @config.host
      port:   @config.port
      secure: false
    }

    # Set up managers.
    if @serverWorkspace?
      @lockManager    = @serverWorkspace.getLockManager(@)
      @fileManager    = @serverWorkspace.getFileManager(@)
      @sessionManager = @serverWorkspace.getSessionManager(@)
      @eventServer    = @serverWorkspace.getEventServer(@)

    # construct router
    @_constructRouter()

    # construct site manager storing meta about all known pages
    @_constructSiteManager()

    # Map actions to proper managers.
    @actionToManagerRouteTable = {}

    if @isDebugMode()
      @log '!!! DEBUG mode enabled !!!'.lightRed

  # Override this
  _constructRouter: () ->
    @router = new K.Router(@)
    @router.addRoute '', 'home'
    @router.addRoute '*url', 'error'

  _constructSiteManager: () ->
    @siteMapManager = new K.SiteMapManager(@)

  isDebugMode: () ->
    return @serverWorkspace.isDebugMode()

  createModel: (uid, done) ->
    err = new K.Error({errorCode:'unknownUid', ext:{uid:uid}})
    done(err, null)

  getConfig: () ->
    return @config

  getLangs: () ->
    return ['en']

  getModel: (uid, done) ->
    if uid of @models
      # Model already loaded. Reuse it.
      model = @models[uid]
      done(null, model)
    else
      # Model not loaded yet. Try load it first time.
      @createModel uid, (err, model) =>
        if not err?
          @models[uid] = model
        done(err, model)

  accessModel: (uid, op, done) ->
    err = new K.Error({errorCode:'unknownUid', ext:{uid:uid, op:op}})
    done(err, null)

  accessModelAndCall: (uid, startOp, finishOp, cb, done) ->
    # TODO: Review.
    @accessModel uid, startOp, (startError, model) =>
      if startError?
        # Error while starting access.
        @appLog.logIfError(startError)
        done(startError)

      else
        # Access started. Do caller work.
        cb model, (cbError, cbResult) =>
          # Work done, finish access.
          @accessModel uid, finishOp, (finishError) =>
            @appLog.logIfError(cbError)
            @appLog.logIfError(finishError)
            done(cbError, cbResult)

  accessArrayOfModelsCommonOp: (arrayOfUids, op, done) ->
    _accessModelWithOp = (uid, done) =>
      @accessModel(uid, op, done)

    async.map arrayOfUids, _accessModelWithOp, (err, arrayOfModels) =>
      if err?
        # Error while accessing one of model.
        done(err)
      else
        # Convert flat [model1, model2, ...] to {uid1:model1, uid2:model2, ...}
        rv = {}
        for idx, uid of arrayOfUids
          rv[uid] = arrayOfModels[idx]

        # Pass result to caller.
        done(null, rv)

  accessArrayOfModels: (uidsAndOps, done) ->
    # TODO: Not implemented.
    # uidsAndOps = {uid1:op1, uid2:op2, ...}
    err = K.Error.createNotImplementedError('KServerApp::accessArrayOfModels')
    done(err)

  getArrayOfModels: (arrayOfUids, done) ->
    async.map arrayOfUids, @getModel.bind(@), (err, arrayOfModels) =>
      if err?
        # Error while getting one of model.
        done(err)
      else
        # Convert flat [model1, model2, ...] to {uid1:model1, uid2:model2, ...}
        rv = {}
        for idx, uid of arrayOfUids
          rv[uid] = arrayOfModels[idx]

        # Pass result to caller.
        done(null, rv)

  getLog: () ->
    return @appLog

  getName: () ->
    return @appName

  _getServerUptimeString: () ->
    rv     = ''
    uptime = (new Date().getTime()) - @serverStartTime

    if uptime < 1000
      rv = uptime + ' ms'

    else
      units = [
        {name: 'days',    weight: 1000 * 60 * 60 * 24}
        {name: 'hours',   weight: 1000 * 60 * 60}
        {name: 'minutes', weight: 1000 * 60}
        {name: 'seconds', weight: 1000}
      ]

      for unit in units
        value = Math.floor(uptime / unit.weight)
        if value > 0
          uptime -= value * unit.weight
          rv += value + ' ' + unit.name + ' '

    return rv.trim()

  _getAppHealth: () ->
    rv = {
      appName: @appName
      serverStats: @serverStats
      serverStartTime: @serverStartTime
      serverUptimeString: @_getServerUptimeString()
    }
    return rv

  getServerStatsCollector: () ->
    return @serverStats

  getAppPath: () ->
    return @appPath

  getBuildPath: () ->
    return @buildPath

  getRootPath: () ->
    return @rootPath

  getRouter: () ->
    return @router

  #
  # Access to managers
  #
  getFileManager: () ->
    return @fileManager

  getLockManager: () ->
    return @lockManager

  getSessionManager: () ->
    return @sessionManager

  getUserManager: () ->
    return @userManager

  getEventServer: () ->
    return @eventServer

  #
  # Validators.
  #
  isOkAsNick: (nick) ->
    return nick? and
    nick.length < NICK_MAX_LEN and
    IS_OK_AS_NICK_REGEXP.test(nick)

  isOkAsEmail: (email) ->
    return email? and
    IS_OK_AS_EMAIL_REGEXP.test(email)

  isOkAsPassword: (password) ->
    return password? and
    password.length < PASSWORD_MAX_LEN and
    password.length >= PASSWORD_MIN_LEN

  #
  # watch
  #
  # Input:
  #   action: 'watch'
  #   items: [
  #     {uid: 'uid.a', stamp: 100},
  #     {uid: 'uid.b', stamp: 50},
  #     {uid: 'uid.c', stamp: 'stamp1'}
  #   ]
  #
  # Output:
  #   status: 'ok'
  #   items: [
  #     {uid: 'uid.a', status: 'watching', sync: 'sync', serverStamp: 100},
  #     {uid: 'uid.b', status: 'watching', sync: 'unknown', serverStamp: null},
  #     {uid: 'uid.c', status: 'watching', sync: 'changed', serverStamp: 'stamp2'},
  #   ]
  #
  # Possible sync statuses:
  # - 'sync'    - browser and server stamps match, browser stamp is up to date (in sync)
  # - 'changed' - browser and server stamps does NOT match, browser stamp is out of date
  # - 'unknown' - server can't resolve stamp for given uid. Browser SHOULD asks server
  #               for given uid manually.
  #
  _doJsonAction_watch: (kRequest, jsonAction, done) ->
    # Check parameters.
    if not jsonAction['items']?
      done null, {status: 'missingParameter', statusHint: 'missing items array'}

    else if not (jsonAction['items'] instanceof Array)
      done null, {status: 'invalidParameter', statusHint: 'items parameter shoulds be an array'}

    else
      # Parameters seem ok.
      @eventServer.watch kRequest, jsonAction.items, (err, items) =>
        if err?
          # Error while setting up watch.
          done(err)

        else
          # Watch set up. Success.
          done null, {status: 'ok', items: items}

  _doJsonAction_test: (kRequest, jsonAction, done) ->
    if not @isDebugMode()
      # WARNING: This is test/development call.
      # Not to be used in production environment.
      done null, {status: 'unhandledAction'}

    else
      switch jsonAction.action
        when 'Test.setUidStamp'
          @eventServer.notifyEvent(jsonAction['uid'], jsonAction['stamp'])
          done null, {status: 'ok'}

        when 'Test.resetEventServer'
          @eventServer.reset()
          done null, {status: 'ok'}

        when 'Test.getEventServerState'
          result = {
            status: 'ok'
            totalNumberOfWatchers: @eventServer.getTotalNumberOfWatchers()
            watchedUids:  @eventServer.getArrayOfWatchedUids()
          }
          done null, result

        when 'Test.getMemoryStats'
          # Force garbage collection explicite.
          try
            global.gc()
          catch e
            console.log("You must run program with 'node --expose-gc index.js' or 'npm start'")
            process.exit()

          # Get number of bytes used current running process.
          result = {
            status: 'ok'
            memoryUsed: process.memoryUsage().heapUsed
          }
          done null, result

        else
          done null, {status: 'unhandledAction'}

  _resolveActionManager: (jsonAction) ->
    rv = null
    # Search route table for given action.
    for uidPrefix, manager of @actionToManagerRouteTable
      if (K.UidUtil.matchPath(jsonAction.action, uidPrefix) != false)
        # Found manager related with given action.
        rv = manager
        break

    return rv

  registerActionManager: (uidPrefix, manager) ->
    if not @actionToManagerRouteTable[uidPrefix]?
      @actionToManagerRouteTable[uidPrefix] = manager

  #
  # Default action processing.
  # This function is called, when any of derived classes cannot handle action.
  # This is last attemp to recognize it as generic built-in action.
  #
  processJsonAction: (kRequest, jsonAction, done) ->
    actionManager = @_resolveActionManager(jsonAction)

    if actionManager?
      # Forward action to one of managers.
      actionManager.doJsonAction(kRequest, jsonAction, done)

    else
      # There is no manager, which can handle action. Try built-in one.
      switch jsonAction.action
        when 'getContent'
          if jsonAction.linkin?
            @createContentResponse(kRequest).getByLinkin jsonAction.linkin, (err, content) ->
              done null, {status: 'ok', linkin: jsonAction.linkin, content: content}
          else
            done null, {status:'missingLinkin'}

        when 'getServerStartTime'
          done null, {status: 'ok', serverStartTime:@serverStartTime}

        when 'log'
          kRequest.getLog().logTime {eventType: 'clientLog', logData: jsonAction.logData}
          done null, {status: 'ok'}

        when 'watch'
          @_doJsonAction_watch(kRequest, jsonAction, done)

        when 'Test.setUidStamp', 'Test.resetEventServer', 'Test.getEventServerState', 'Test.getMemoryStats'
          @_doJsonAction_test(kRequest, jsonAction, done)

        else
          done null, {status: 'unhandledAction'}

  _processJsonActionsInternal: (kRequest, jsonActions, cb) ->
    if (jsonActions instanceof Array)

      # Wrapper function to add kRequest parameter to every call
      # to @processJsonAction.
      _processJsonActionWithKRequest = (jsonAction, cb) =>
        @processJsonAction kRequest, jsonAction, (err, actionResult) =>
          @serverStats.reportJsonAction(actionResult)
          if err?
            # Generic, per action error.
            @appLog.logTime {eventType: 'actionInternalError', err:err}
            cb null, {status: 'internalError'}
          else
            cb null, actionResult

      async.mapSeries jsonActions, _processJsonActionWithKRequest, (err, asyncResult) =>
        if err?
          # Generic process actions fail. This is error common to every
          # processed actions if more than one were processed.
          @appLog.logTime {eventType: 'actionProcessError', err:err}
          cb({status:'internalError'})
        else
          cb(asyncResult)
    else
      cb({status:'invalid action list'})

  processJsonActions: (kRequest, jsonActions, cb) ->
    # if we're testing json-dead, let's just never call the cb here :)
    if @config.testJsonActionsDead
      return

    # if we're testing json-slow, let's just wait here
    if @config.testJsonActionsSlowMs > 0
      setTimeout( () =>
        @_processJsonActionsInternal(kRequest, jsonActions, cb)
      , @config.testJsonActionsSlowMs
      )
    else
      @_processJsonActionsInternal(kRequest, jsonActions, cb)

  postProcessJsonAction: (kRequest, jsonAction, jsonResponse, callback) ->
    callback null

  postProcessJsonActions: (kRequest, jsonActions, jsonResponses) ->
    # Validate input
    if jsonActions instanceof Array and
    jsonResponses instanceof Array and
    jsonActions.length is jsonResponses.length
      # Pack actions and responses into one array
      # as it's easier to pass to async.map
      jsonActionsAndResponses = []
      idx = 0

      for idx of jsonResponses
        jsonActionsAndResponses.push
          jsonAction:   jsonActions[idx]
          jsonResponse: jsonResponses[idx]

      # Internal wrapper for postProcessJsonAction(). We need it beacuse:
      # - it splits {action, response} struct into separate parameters
      # - it adds common kRequest parameter
      _postProcessJsonAction = (jsonActionAndResponse, done) =>
        jsonAction   = jsonActionAndResponse.jsonAction
        jsonResponse = jsonActionAndResponse.jsonResponse
        @postProcessJsonAction(kRequest, jsonAction, jsonResponse, done)

      # Post process actions one by one.
      async.mapSeries jsonActionsAndResponses, _postProcessJsonAction, null

  processWsMsg: (ws, msg) ->
      # try parsing json
      try
        actionPack = JSON.parse(msg)
      catch error
        actionPack = null

      # play with actionPack
      if actionPack? and actionPack.actionPackId?
        ws.upgradeReq.log.logTime {
          eventType: 'Websocket:actionPackReceived'
          wsId: ws.id
          actionPackId: actionPack.actionPackId
        }

        @processJsonActions ws.upgradeReq.kRequest, actionPack.actions, (jsonResponses) =>
          ws.upgradeReq.log.logTime {
            eventType: 'Websocket:actionPackProcessed'
            wsId: ws.id
            actionPackId: actionPack.actionPackId
          }

          response = {actionPackId:actionPack.actionPackId, actionsResults:jsonResponses}

          ws.trySend JSON.stringify(response)
      else
        ws.trySend JSON.stringify({status:'syntax error'})

  onGotWsConnection: (ws) ->
    # log it
    @appLog.logTime {
      eventType: 'Websocket:gotConnection'
      host:      ws.upgradeReq.headers.host
      wsId:      ws.id
    }
    @serverStats.reportWebsocketOpen()

    # attach myself
    ws.app = @

  onLostWsConnection: (ws) ->
    # log it
    ws.upgradeReq.log.logTime {
      eventType: 'Websocket:lostConnection'
      host:      ws.upgradeReq.headers.host
      wsId:      ws.id
    }
    @serverStats.reportWebsocketClose()

  _createKRequestResponseConfig: (kRequest, done) ->
    done(null, @defaultResponseConfig)

  _createKRequestObject: (expressReq, expressRes, done) ->
    done(null, new K.Request(@))

  createContentResponse: (kRequest) ->
    return new K.ServerContentResponse(kRequest)

  createKRequest: (expressReq, expressRes, done) ->
    @_createKRequestObject expressReq, expressRes, (err, newKRequest) =>
      expressReq.kRequest = newKRequest
      expressReq.kRequest.attachExpressRequestAndResponse expressReq, expressRes, (err) =>
        @_createKRequestResponseConfig expressReq.kRequest, (err, responseConfig) =>
          # we ignore err here
          expressReq.kRequest.setResponseConfig responseConfig, (err) =>
            done(null, expressReq.kRequest)

  _startTheListeningAsNonVHost: (port) ->
    @appLog.log 'Server starting as non-vhost on port:' + port
    httpServer = @express.listen port, () =>
      host = httpServer.address().address
      port = httpServer.address().port
      @log 'HTTP server listening at http://' + host + ':' + port

      # WebSockets Server
      if not @config.testJsonActionsDead and @config.useWebSocketServer
        @wsServer = new K.WebSocketServer(@, {server: httpServer})

    # request statistics
    K.ServerStatsCollector.setRequestStatsHandler(httpServer, @serverStats)

  _createLogContextForNewRequest: (app, req) ->
    if @appLog.isFake()
      req.log = reqLog = new K.LogCtxFake()
    else
      # start request log
      reqLog = new K.LogCtxRequest(app.appLog, req)

      # check for request log made at vHost level
      if req.log?
        # make logical connection between vHost request and app request
        req.logVHost = req.log
        req.logVHost.log {event:'vHost request catched by app', ctxId: reqLog.getCtxStamp()}

        # Notice: do not call logRequest(), as this is already logged at vHost level
        req.log = reqLog
      else
        # register request logger and log default stuff (headers, client ip, time, etc.)
        req.log = reqLog
        reqLog.logRequest()

  _getRequestRobotsTxtAsString: (kRequest, done) ->
    done(null, 'Sitemap: ' + kRequest.getRootLink('sitemap.xml'))

  _getRequestRobotsTxt: (kRequest, res) ->
    @_getRequestRobotsTxtAsString kRequest, (err, robotsTxtAsString) =>
      if err?
        res.status(404).send('')
      else
        res.send(robotsTxtAsString)

  _getRequestSitemapXmlAsString: (kRequest, done) ->
    if @siteMapManager?
      @siteMapManager.getSiteMapXmlAsString(kRequest, done)
    else
     done new K.Error {errorCode: 'sitemapManagerNotFound'}

  _getRequestSitemapXml: (kRequest, res) ->
    @_getRequestSitemapXmlAsString kRequest, (err, sitemapXmlAsString) =>
      if err?
        res.status(404).send('')
      else
        res.set('Content-Type', 'text/xml')
        res.send(sitemapXmlAsString)

  start: (done) ->
    @serverStartTime = (new Date()).getTime()
    @express = express()

    @express.use (req, res, next) =>
      @_createLogContextForNewRequest(@, req)
      # pass to next express-MW
      next()

    # express middlewares
    @express.use(bodyParser.json())
    @express.use(connectRestreamer())
    @express.use(compression())

    # vhost configs
    if (not @config.virtualHosts?) or (@config.virtualHosts == 0)
      @_startTheListeningAsNonVHost(@config.port)
    else
      @appLog.log 'HTTP server working in virtual host mode'

    #
    # Default routings
    #
    @express.post '/jsonActions', (req, res) =>
      req.log.logTime 'POST: jsonActions requested'

      # Init and inject K.Request object into express object.
      @createKRequest req, res, (err, kRequest) =>

        # Process actions list received inside request body
        @processJsonActions kRequest, req.body, (jsonResponses) ->
          req.log.logTime 'POST: jsonActions responses ready'

          # Append sid to cookie if needed.
          if kRequest.getKemuSid()?
            sid    = kRequest.getKemuSid()
            cookie = 'sid=' + sid + ';path=/'
            res.append('set-cookie', cookie)

          res.json(jsonResponses)
          req.log.logTime 'POST: jsonActions sent'

    @express.get '/favicon.ico', (req, res) =>
      req.log.logTime 'favicon'
      res.status(404).send 'no fav-icon :P'

    @express.get '/_health', (req, res) =>
      @nateRoot = new K.NateHtmlDocument()
      @nateRoot.body().newH1('Server health')

      appHealth = @_getAppHealth()
      nTable    = @nateRoot.body().newTable()

      _putTwoColumnsRow = (label, value) ->
        nRow = nTable.newTr()
        nRow.newTd(label)
        nRow.newTd(value)

      startTimeObj = new Date()
      startTimeObj.setTime(appHealth.serverStartTime)
      startTimeString = startTimeObj.toUTCString()

      _putTwoColumnsRow('App name'       , appHealth.appName)
      _putTwoColumnsRow('App start time' , startTimeString)
      _putTwoColumnsRow('App uptime'     , appHealth.serverUptimeString)
      _putTwoColumnsRow('POST hits'      , appHealth.serverStats.getPostHitsCounter())
      _putTwoColumnsRow('GET hits'       , appHealth.serverStats.getGetHitsCounter())
      _putTwoColumnsRow('JSON/OK hits'   , appHealth.serverStats.getJsonOkCounter())
      _putTwoColumnsRow('JSON/ERROR hits', appHealth.serverStats.getJsonErrorCounter())
      _putTwoColumnsRow('Open websockets', appHealth.serverStats.getOpenWebsocketsCounter())

      res.status(200).send(@nateRoot.render())

    @express.get '/robots.txt', (req, res) =>
      @createKRequest req, res, (err, kRequest) =>
        @_getRequestRobotsTxt(kRequest, res)

    @express.get '/sitemap.xml', (req, res) =>
      @createKRequest req, res, (err, kRequest) =>
        @_getRequestSitemapXml(kRequest, res)

    # Server started.
    done?(null)

  getExpress: () ->
    return @express

  log: (msg...) ->
    args = Array.prototype.slice.call(arguments)
    args.splice 0, 0, '[', @appName, ']'
    console.log.apply @, args

  logDebug: (msg...) ->
    if @isDebugMode()
      args = Array.prototype.slice.call(arguments)
      args.splice 0, 0, '[', @appName, ']'
      console.log.apply @, args
