K = require('kcore')

class K.SiteMapManager
  constructor: (@app) ->
    # Check parameters.
    K.Error.throwExceptionIfParameterNotSet(@app, 'app')
    K.Error.throwExceptionIfParameterNotSet(@app.getLog(), 'appLog')
    K.Error.throwExceptionIfParameterNotSet(@app.getFileManager(), 'fileManager')

    # Set up object.
    @siteMap           = {}
    @siteMapByLang     = {}
    @siteMapForBrowser = {}

    # Init maps to store filtered site maps for different languages.
    for langId in @app.getLangs()
      @siteMapByLang[langId]     = {}
      @siteMapForBrowser[langId] = {}

    @appLog      = @app.getLog()
    @fileManager = @app.getFileManager()

  start: (done) ->
    done(null)

  _isOkAsPageMeta: (meta) ->
    rv = false
    if meta? and meta.linkin?
      rv = true
      for langId, linkinForGivenLang of meta.linkin
        if (not linkinForGivenLang.length) or (linkinForGivenLang.length is 0)
          # Error, one of linkin entry is not array or array is empty.
          rv = false
          break

    return rv

  _filterPageMetaByLang: (pageMeta, langId) ->
    rv = {}
    for key, value of pageMeta
      if value[langId]?
        # Multilanguage field. Get version for selected language only.
        rv[key] = {}
        rv[key][langId] = value[langId]

      else
        # Field is independent on langauge. Get whole value.
        rv[key] = value

    # Special case - always get linkin for all languages.
    rv.linkin = pageMeta.linkin

    # Pass filtered result to caller.
    return rv

  _filterPageMetaForBrowser: (pageMeta, langId) ->
    if pageMeta.navigationTitle?
      navigationTitle = pageMeta.navigationTitle[langId]
    else
      navigationTitle = pageMeta.title[langId]

    rv = {
      linkin:          pageMeta.linkin
      navigationTitle: navigationTitle
    }
    return rv

  addPage: (pageId, meta) ->
    if @_isOkAsPageMeta(meta)
      # Add page meta to common site map.
      @siteMap[pageId] = meta
      @siteMap[pageId].pageId = pageId

      # Prepare filtered versions of page meta.
      for langId in @app.getLangs()
        # Prepare page meta filtered by language.
        @siteMapByLang[langId][pageId] = @_filterPageMetaByLang(meta, langId)
        # Prepare page meta ready to send to browser.
        @siteMapForBrowser[langId][pageId] = @_filterPageMetaForBrowser(meta, langId)

    else
      # Something wrong with page meta.
      err = new K.Error({errorCode:'invalidPageMeta', ext:{pageId: pageId, meta: meta}})
      @appLog.logIfError(err)
      throw err

  addPagesFromJsonFile: (filename, done) ->
    @fileManager.loadDataFromJson filename, true, (err, data) =>
      if not err?
        for pageId, pageData of data
          @addPage(pageId, pageData)
      done(err)

  getPageMeta: (pageId) ->
    return @siteMap[pageId]

  getPageProperty: (pageId, propertyId, langId = null) ->
    if langId?
      rv = @siteMap[pageId]?[propertyId]?[langId]
    else
      rv = @siteMap[pageId]?[propertyId]
    return rv

  findPageByLinkin: (linkin, langId = null) ->
    # TODO: Optimize this.
    rv = null

    if langId?
      # Caller passed language id. Search one specific language only.
      for pageId, pageMeta of @siteMap
        if pageMeta.linkin?
          if linkin in pageMeta.linkin[langId]
            rv = pageMeta
            break
    else
      # Language not specified. Search linkins for all known languages.
      for pageId, pageMeta of @siteMap
        if rv?
          # Page found, don't search longer.
          break
        else
          # Page not found yet, go on.
          if pageMeta.linkin?
            for langId, linkinForGivenLang of pageMeta.linkin
              if linkin in linkinForGivenLang
                rv = pageMeta
                break

    # Pass result to caller.
    return rv

  findPageIdByLinkin: (linkin, langId = null) ->
    page = @findPageByLinkin(linkin, langId)
    return page?.pageId

  getDefaultLinkinsForAlternativeLangs: (pageId) ->
    if @siteMap[pageId]?
      rv = {}
      for langId, linkinsForGivenLang of @siteMap[pageId].linkin
        rv[langId] = linkinsForGivenLang[0]
    return rv

  # Example input: {pageId: 'news', langId: 'pl'}
  # Example output: 'aktualnosci'
  getCanonicalLinkin: (params) ->
    if params? and
    params.pageId? and
    params.langId?
      # All needed parameters passed, go on.
      rv = @siteMap[params.pageId]?.linkin[params.langId]?[0]

    return rv

  getSiteMapForForLang: (langId) ->
    return @siteMapByLang[langId]

  getSiteMapForBrowser: (kRequest) ->
    return @siteMapForBrowser[kRequest.getLangId()]

  isPageExistentById: (pageId) ->
    return @siteMap[pageId]?

  _buildSiteMapXml_addRegisteredPages: (kRequest, done) ->
    rv = ''

    langId = kRequest.getLangId?()
    if not langId?
      langId = 'en'

    for id, page of @siteMap
      rv += '<url>' + "\n"
      rv += '  <loc>' + kRequest.getRootLink(page.linkin[langId][0]) + '</loc>' + "\n"
      rv += '</url>' + "\n"

    done(null, rv)

  _buildSiteMapXml_addAdditionalPages: (kRequest, done) ->
    # Override this to handle custom pages like calculators
    done(null, '')

  getSiteMapXmlAsString: (kRequest, done) ->
    @_buildSiteMapXml_addRegisteredPages kRequest, (err, registeredPagesMap) =>
      @_buildSiteMapXml_addAdditionalPages kRequest, (err, additionalPagesMap) =>
        rv  = '<?xml version="1.0" encoding="UTF-8"?>' + "\n"
        rv += '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">' + "\n"
        rv += registeredPagesMap
        rv += additionalPagesMap
        rv += '</urlset>' + "\n"

        if rv.indexOf('&') isnt -1
          # Throw exception if not escaped chars detected.
          err = new K.Error {errorCode: 'NotEscapedSiteMap'}
          @appLog.logIfError(err)
          done(err)
          throw(err)

        else
          # Success, return final sitemap to caller.
          done(null,  rv)
